import { Request, Response } from "express";
import { User } from "../models/user";
import {
  saveSession,
  hashPassword,
  decodeBase64,
  matchPassword,
  getUserSession,
  getUser,
} from "../utils/utils";
import { isUsernameValid, isPasswordValid } from "../validation/validation";
import { Session } from "../models/session";
import { MONGODB_IGNORED_FIELDS_USER } from "../utils/constants";

export async function SignUp(req: Request, res: Response) {
  try {
    let { AccountName, Username, Password } = req.body;

    if (!AccountName || !Username || !Password) {
      return res.status(400).json({ status: "err" });
    }

    AccountName = AccountName.toLowerCase();

    const existingUser = await User.findOne({ accountName: AccountName })
      .select(`accountName -_id`)
      .lean();

    if (existingUser) {
      return res.status(400).json({ status: 1 });
    }

    const decodedPassword = decodeBase64(Password);

    if (!isPasswordValid(decodedPassword)) {
      return res.status(400).json({ status: "err" });
    }

    const hashedPassword = await hashPassword(decodedPassword);

    const user = new User({
      accountName: AccountName,
      username: Username,
      password: hashedPassword,
    });

    user
      .save()
      .then(() => saveSession(res, user.userId, Username))
      .catch((err) => {
        console.error("error on signup saving session:", err);
        res.status(500).json({ status: "err" });
      });
  } catch (error) {
    console.error("error on signup:", error);
    res.status(500).json({ status: "err" });
  }
}

export async function Login(req: Request, res: Response) {
  try {
    const { AccountName, Password } = req.body;

    if (!AccountName || !Password) {
      return res.status(400).json({ status: "err" });
    }

    const user = await User.findOne({ accountName: AccountName })
      .select(`userId accountName username password -_id`)
      .lean();

    if (!user) {
      return res.status(401).json({ status: "err" });
    }

    if (
      user.accountName === null ||
      user.accountName === undefined ||
      user.password === undefined ||
      user.password === null
    ) {
      return res.status(401).json({ status: "err" });
    }

    const decodedPassword = decodeBase64(Password);
    const isPasswordValid = await matchPassword(decodedPassword, user.password);

    if (!isPasswordValid) {
      return res.status(401).json({ status: "err" });
    }

    saveSession(res, user.userId, user.username);
  } catch (error) {
    console.error("error on login:", error);
    res.status(500).json({ status: "err" });
  }
}

export async function ChangeUsername(req: Request, res: Response) {
  try {
    const { NewUsername } = req.body;

    if (!NewUsername || !isUsernameValid(NewUsername)) {
      return res.status(400).json({ status: "err" });
    }

    const session = await getUserSession(req, "userId");

    if (!session) {
      return res.status(401).json({ status: "err" });
    }

    await User.updateOne(
      { userId: session.userId },
      { $set: { username: NewUsername } }
    );

    res.status(200).json({ status: "ok" });
  } catch (error) {
    console.error("error on changing username", error);
    res.status(500).json({ status: "err" });
  }
}

export async function ChangePassword(req: Request, res: Response) {
  try {
    // CP = current password
    // NP = new password
    const { CP, NP } = req.body;

    if (!CP || !NP) {
      return res.status(400).json({ status: "err" });
    }

    const session = await getUserSession(req, "userId");

    if (!session || !session.userId) {
      return res.status(401).json({ status: "err" });
    }

    const decodedCurrentPassword = decodeBase64(CP);

    if (!isPasswordValid(decodedCurrentPassword)) {
      return res.status(400).json({ status: "err" });
    }

    const userId = session.userId;

    const existingUser = await getUser(userId);

    if (!existingUser) {
      return res.status(400).json({ status: 1 });
    }

    const isPasswordMatching = await matchPassword(
      decodedCurrentPassword,
      existingUser.password
    );

    const decodedNewPassword = decodeBase64(NP);

    if (!isPasswordMatching || !isPasswordValid(decodedNewPassword)) {
      return res.status(401).json({ status: "err" });
    }

    await User.updateOne(
      { userId: userId },
      { $set: { password: await hashPassword(decodedNewPassword) } }
    );

    res.status(200).json({ status: "ok" });

    // delete all sessions
    await Session.deleteMany({ userId: userId });
  } catch (error) {
    console.error("error on changing password", error);
    res.status(500).json({ status: "err" });
  }
}

export async function Logout(req: Request, res: Response) {
  try {
    const session = await getUserSession(req);

    if (!session) {
      return res.status(401).json({ status: "err" });
    }

    await Session.deleteOne({ sessionId: session.sessionId });

    res.status(200).json({ status: "ok" });
  } catch (error) {
    console.error("error on logout", error);
    res.status(500).json({ status: "err" });
  }
}

export async function IsAccountNameAvailable(req: Request, res: Response) {
  try {
    let { accountName } = req.params;

    accountName = accountName.toLowerCase();

    if (!accountName) {
      return res.status(400).json({ status: "err" });
    }

    const existingUser = await User.findOne({ accountName: accountName })
      .select(`accountName -_id`)
      .lean();

    if (existingUser) {
      return res.status(422).json({ status: "not found" });
    }

    res.status(200).json({ status: "ok" });
  } catch (error) {
    console.error("error on checking account name availability", error);
    res.status(500).json({ status: "err" });
  }
}

// appstart
export async function GetUserProfile(req: Request, res: Response) {
  try {
    const session = await getUserSession(req, "userId");

    if (!session || !session.userId) {
      return res.status(401).json({ status: "err" });
    }

    const user = await getUser(
      session.userId,
      `${MONGODB_IGNORED_FIELDS_USER} -createdAt`
    );

    if (!user) {
      return res.status(404).json({ status: "err" });
    }

    res.json(user);
  } catch (error) {
    console.error("error on get user profile:", error);
    res.status(500).json({ status: "err" });
  }
}

export async function UpdateAvatar(req: Request, res: Response) {
  try {
    console.log("req", req.file, req.body);
  } catch (error) {
    console.error("error on get user profile:", error);
    res.status(500).json({ status: "err" });
  }
}
